package javax.swing.text.rtf;

import java.util.Vector;
import javax.swing.text.BadLocationException;
import javax.swing.text.TabStop;
import java.util.Enumeration;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import java.util.Hashtable;
import java.io.IOException;
import javax.swing.text.Element;
import javax.swing.text.Document;
import java.awt.Color;
import javax.swing.text.Segment;
import javax.swing.text.MutableAttributeSet;
import java.io.OutputStream;
import javax.swing.text.AttributeSet;
import java.util.Dictionary;

class RTFGenerator
{
    Dictionary<Object, Integer> colorTable;
    int colorCount;
    Dictionary<String, Integer> fontTable;
    int fontCount;
    Dictionary<AttributeSet, Integer> styleTable;
    int styleCount;
    OutputStream outputStream;
    boolean afterKeyword;
    MutableAttributeSet outputAttributes;
    int unicodeCount;
    private Segment workingSegment;
    int[] outputConversion;
    public static final Color defaultRTFColor;
    public static final float defaultFontSize = 12.0f;
    public static final String defaultFontFamily = "Helvetica";
    private static final Object MagicToken;
    protected static RTFGenerator.CharacterKeywordPair[] textKeywords;
    static final char[] hexdigits;
    
    public static void writeDocument(Document document, OutputStream outputStream) throws IOException {
        RTFGenerator rtfGenerator = new RTFGenerator(outputStream);
        Element defaultRootElement = document.getDefaultRootElement();
        rtfGenerator.examineElement(defaultRootElement);
        rtfGenerator.writeRTFHeader();
        rtfGenerator.writeDocumentProperties(document);
        for (int elementCount = defaultRootElement.getElementCount(), i = 0; i < elementCount; ++i) {
            rtfGenerator.writeParagraphElement(defaultRootElement.getElement(i));
        }
        rtfGenerator.writeRTFTrailer();
    }
    
    public RTFGenerator(OutputStream outputStream) {
        (this.colorTable = new Hashtable()).put(RTFGenerator.defaultRTFColor, Integer.valueOf(0));
        this.colorCount = 1;
        this.fontTable = new Hashtable();
        this.fontCount = 0;
        this.styleTable = new Hashtable();
        this.styleCount = 0;
        this.workingSegment = new Segment();
        this.outputStream = outputStream;
        this.unicodeCount = 1;
    }
    
    public void examineElement(Element element) {
        AttributeSet attributes = element.getAttributes();
        this.tallyStyles(attributes);
        if (attributes != null) {
            Color foreground = StyleConstants.getForeground(attributes);
            if (foreground != null && this.colorTable.get(foreground) == null) {
                this.colorTable.put(foreground, new Integer(this.colorCount));
                ++this.colorCount;
            }
            Object attribute = attributes.getAttribute(StyleConstants.Background);
            if (attribute != null && this.colorTable.get(attribute) == null) {
                this.colorTable.put(attribute, new Integer(this.colorCount));
                ++this.colorCount;
            }
            String fontFamily = StyleConstants.getFontFamily(attributes);
            if (fontFamily == null) {
                fontFamily = "Helvetica";
            }
            if (fontFamily != null && this.fontTable.get(fontFamily) == null) {
                this.fontTable.put(fontFamily, new Integer(this.fontCount));
                ++this.fontCount;
            }
        }
        for (int elementCount = element.getElementCount(), i = 0; i < elementCount; ++i) {
            this.examineElement(element.getElement(i));
        }
    }
    
    private void tallyStyles(AttributeSet resolveParent) {
        while (resolveParent != null) {
            if (resolveParent instanceof Style && (Integer)this.styleTable.get(resolveParent) == null) {
                ++this.styleCount;
                this.styleTable.put(resolveParent, new Integer(this.styleCount));
            }
            resolveParent = resolveParent.getResolveParent();
        }
    }
    
    private Style findStyle(AttributeSet resolveParent) {
        while (resolveParent != null) {
            if (resolveParent instanceof Style && this.styleTable.get(resolveParent) != null) {
                return (Style)resolveParent;
            }
            resolveParent = resolveParent.getResolveParent();
        }
        return null;
    }
    
    private Integer findStyleNumber(AttributeSet resolveParent, String s) {
        while (resolveParent != null) {
            if ((resolveParent instanceof Style)) {
                Integer n = (Integer)this.styleTable.get(resolveParent);
                if (n != null && (s == null || s.equals(resolveParent.getAttribute("style:type")))) {
                    return n;
                }
            }
            resolveParent = resolveParent.getResolveParent();
        }
        return null;
    }
    
    private static Object attrDiff(MutableAttributeSet set, AttributeSet set2, Object o, Object o2) {
        Object attribute = set.getAttribute(o);
        Object attribute2 = set2.getAttribute(o);
        if (attribute2 == attribute) {
            return null;
        }
        if (attribute2 == null) {
            set.removeAttribute(o);
            if (o2 != null && !o2.equals(attribute)) {
                return o2;
            }
            return null;
        }
        else {
            if (attribute == null || !equalArraysOK(attribute, attribute2)) {
                set.addAttribute(o, attribute2);
                return attribute2;
            }
            return null;
        }
    }
    
    private static boolean equalArraysOK(Object o, Object obj) {
        if (o == obj) {
            return true;
        }
        if (o == null || obj == null) {
            return false;
        }
        if (o.equals(obj)) {
            return true;
        }
        if (!o.getClass().isArray() || !obj.getClass().isArray()) {
            return false;
        }
        Object[] array = (Object[])(Object[])o;
        Object[] array2 = (Object[])(Object[])obj;
        if (array.length != array2.length) {
            return false;
        }
        int length = array.length;
        int i = 0;
        while (i < length) {
            if (!equalArraysOK(array[i], array2[i])) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    public void writeLineBreak() throws IOException {
        this.writeRawString("\n");
        this.afterKeyword = false;
    }
    
    public void writeRTFHeader() throws IOException {
        this.writeBegingroup();
        this.writeControlWord("rtf", 1);
        this.writeControlWord("ansi");
        this.outputConversion = outputConversionForName("ansi");
        this.writeLineBreak();
        String[] array = new String[this.fontCount];
        Enumeration<String> keys = this.fontTable.keys();
        while (keys.hasMoreElements()) {
            String s = (String)keys.nextElement();
            array[(int)this.fontTable.get(s)] = s;
        }
        this.writeBegingroup();
        this.writeControlWord("fonttbl");
        for (int i = 0; i < this.fontCount; ++i) {
            this.writeControlWord("f", i);
            this.writeControlWord("fnil");
            this.writeText(array[i]);
            this.writeText(";");
        }
        this.writeEndgroup();
        this.writeLineBreak();
        if (this.colorCount > 1) {
            Color[] array2 = new Color[this.colorCount];
            Enumeration<Object> keys2 = this.colorTable.keys();
            while (keys2.hasMoreElements()) {
                Color color = (Color)keys2.nextElement();
                array2[(int)this.colorTable.get(color)] = color;
            }
            this.writeBegingroup();
            this.writeControlWord("colortbl");
            for (int j = 0; j < this.colorCount; ++j) {
                Color color2 = array2[j];
                if (color2 != null) {
                    this.writeControlWord("red", color2.getRed());
                    this.writeControlWord("green", color2.getGreen());
                    this.writeControlWord("blue", color2.getBlue());
                }
                this.writeRawString(";");
            }
            this.writeEndgroup();
            this.writeLineBreak();
        }
        if (this.styleCount > 1) {
            this.writeBegingroup();
            this.writeControlWord("stylesheet");
            Enumeration<AttributeSet> keys3 = this.styleTable.keys();
            while (keys3.hasMoreElements()) {
                Style style = (Style)keys3.nextElement();
                int intValue = (int)this.styleTable.get(style);
                this.writeBegingroup();
                String s2 = (String)style.getAttribute("style:type");
                if (s2 == null) {
                    s2 = "paragraph";
                }
                if (s2.equals("character")) {
                    this.writeControlWord("*");
                    this.writeControlWord("cs", intValue);
                }
                else if (s2.equals("section")) {
                    this.writeControlWord("*");
                    this.writeControlWord("ds", intValue);
                }
                else {
                    this.writeControlWord("s", intValue);
                }
                AttributeSet resolveParent = style.getResolveParent();
                Object o;
                if (resolveParent == null) {
                    o = new SimpleAttributeSet();
                }
                else {
                    o = new SimpleAttributeSet(resolveParent);
                }
                this.updateSectionAttributes(o, style, false);
                this.updateParagraphAttributes(o, style, false);
                this.updateCharacterAttributes(o, style, false);
                AttributeSet resolveParent2 = style.getResolveParent();
                if (resolveParent2 != null && resolveParent2 instanceof Style) {
                    Integer n = (Integer)this.styleTable.get(resolveParent2);
                    if (n != null) {
                        this.writeControlWord("sbasedon", (int)n);
                    }
                }
                Style style2 = (Style)style.getAttribute("style:nextStyle");
                if (style2 != null) {
                    Integer n2 = (Integer)this.styleTable.get(style2);
                    if (n2 != null) {
                        this.writeControlWord("snext", (int)n2);
                    }
                }
                Boolean b = (Boolean)style.getAttribute("style:hidden");
                if (b != null && (boolean)b) {
                    this.writeControlWord("shidden");
                }
                Boolean b2 = (Boolean)style.getAttribute("style:additive");
                if (b2 != null && (boolean)b2) {
                    this.writeControlWord("additive");
                }
                this.writeText(style.getName());
                this.writeText(";");
                this.writeEndgroup();
            }
            this.writeEndgroup();
            this.writeLineBreak();
        }
        this.outputAttributes = new SimpleAttributeSet();
    }
    
    void writeDocumentProperties(Document document) throws IOException {
        boolean n = false;
        for (int i = 0; i < RTFAttributes.attributes.length; ++i) {
            RTFAttribute rtfAttribute = RTFAttributes.attributes[i];
            if (rtfAttribute.domain() == 3 && rtfAttribute.writeValue(document.getProperty(rtfAttribute.swingName()), this, false)) {
                n = true;
            }
        }
        if (!(n == 0)) {
            this.writeLineBreak();
        }
    }
    
    public void writeRTFTrailer() throws IOException {
        this.writeEndgroup();
        this.writeLineBreak();
    }
    
    protected void checkNumericControlWord(MutableAttributeSet set, AttributeSet set2, Object o, String s, float floatValue, float n) throws IOException {
        Object attrDiff;
        if ((attrDiff = attrDiff(set, set2, o, RTFGenerator.MagicToken)) != null) {
            if (attrDiff == RTFGenerator.MagicToken) {
            }
            else {
                floatValue = ((Number)attrDiff).floatValue();
            }
            this.writeControlWord(s, Math.round(floatValue * n));
        }
    }
    
    protected void checkControlWord(MutableAttributeSet set, AttributeSet set2, RTFAttribute rtfAttribute) throws IOException {
        Object attrDiff;
        if ((attrDiff = attrDiff(set, set2, rtfAttribute.swingName(), RTFGenerator.MagicToken)) != null) {
            if (attrDiff == RTFGenerator.MagicToken) {
                attrDiff = null;
            }
            rtfAttribute.writeValue(attrDiff, this, true);
        }
    }
    
    protected void checkControlWords(MutableAttributeSet set, AttributeSet set2, RTFAttribute[] array, int n) throws IOException {
        for (RTFAttribute rtfAttribute : array) {
            if (rtfAttribute.domain() == n) {
                this.checkControlWord(set, set2, rtfAttribute);
            }
        }
    }
    
    void updateSectionAttributes(MutableAttributeSet set, AttributeSet set2, boolean b) throws IOException {
        if (b) {
            Object attribute = set.getAttribute("sectionStyle");
            Integer styleNumber = this.findStyleNumber(set2, "section");
            if (attribute != styleNumber) {
                if (attribute != null) {
                    this.resetSectionAttributes(set);
                }
                if (styleNumber != null) {
                    this.writeControlWord("ds", (int)styleNumber);
                    set.addAttribute("sectionStyle", styleNumber);
                }
                else {
                    set.removeAttribute("sectionStyle");
                }
            }
        }
        this.checkControlWords(set, set2, RTFAttributes.attributes, 2);
    }
    
    protected void resetSectionAttributes(MutableAttributeSet default1) throws IOException {
        this.writeControlWord("sectd");
        for (int length = RTFAttributes.attributes.length, i = 0; i < length; ++i) {
            RTFAttribute rtfAttribute = RTFAttributes.attributes[i];
            if (rtfAttribute.domain() == 2) {
                rtfAttribute.setDefault(default1);
            }
        }
        default1.removeAttribute("sectionStyle");
    }
    
    void updateParagraphAttributes(MutableAttributeSet set, AttributeSet set2, boolean b) throws IOException {
        Object attribute;
        Integer styleNumber;
        if (b) {
            attribute = set.getAttribute("paragraphStyle");
            styleNumber = this.findStyleNumber(set2, "paragraph");
            if (attribute != styleNumber && attribute != null) {
                this.resetParagraphAttributes(set);
                attribute = null;
            }
        }
        else {
            attribute = null;
            styleNumber = null;
        }
        Object attribute2 = set.getAttribute("tabs");
        Object attribute3 = set2.getAttribute("tabs");
        if (attribute2 != attribute3 && attribute2 != null) {
            this.resetParagraphAttributes(set);
            attribute2 = null;
            attribute = null;
        }
        if (attribute != styleNumber && styleNumber != null) {
            this.writeControlWord("s", (int)styleNumber);
            set.addAttribute("paragraphStyle", styleNumber);
        }
        this.checkControlWords(set, set2, RTFAttributes.attributes, 1);
        if (attribute2 != attribute3 && attribute3 != null) {
            TabStop[] array = (TabStop[])(TabStop[])attribute3;
            for (int i = 0; i < array.length; ++i) {
                TabStop tabStop = array[i];
                switch (tabStop.getAlignment()) {
                    case 0:
                    case 5: {
                        break;
                    }
                    case 1: {
                        this.writeControlWord("tqr");
                        break;
                    }
                    case 2: {
                        this.writeControlWord("tqc");
                        break;
                    }
                    case 4: {
                        this.writeControlWord("tqdec");
                        break;
                    }
                }
                switch (tabStop.getLeader()) {
                    case 0: {
                        break;
                    }
                    case 1: {
                        this.writeControlWord("tldot");
                        break;
                    }
                    case 2: {
                        this.writeControlWord("tlhyph");
                        break;
                    }
                    case 3: {
                        this.writeControlWord("tlul");
                        break;
                    }
                    case 4: {
                        this.writeControlWord("tlth");
                        break;
                    }
                    case 5: {
                        this.writeControlWord("tleq");
                        break;
                    }
                }
                int round = Math.round(20.0f * tabStop.getPosition());
                if (tabStop.getAlignment() == 5) {
                    this.writeControlWord("tb", round);
                }
                else {
                    this.writeControlWord("tx", round);
                }
            }
            set.addAttribute("tabs", array);
        }
    }
    
    public void writeParagraphElement(Element element) throws IOException {
        this.updateParagraphAttributes(this.outputAttributes, element.getAttributes(), true);
        for (int elementCount = element.getElementCount(), i = 0; i < elementCount; ++i) {
            this.writeTextElement(element.getElement(i));
        }
        this.writeControlWord("par");
        this.writeLineBreak();
    }
    
    protected void resetParagraphAttributes(MutableAttributeSet default1) throws IOException {
        this.writeControlWord("pard");
        default1.addAttribute(StyleConstants.Alignment, Integer.valueOf(0));
        for (int length = RTFAttributes.attributes.length, i = 0; i < length; ++i) {
            RTFAttribute rtfAttribute = RTFAttributes.attributes[i];
            if (rtfAttribute.domain() == 1) {
                rtfAttribute.setDefault(default1);
            }
        }
        default1.removeAttribute("paragraphStyle");
        default1.removeAttribute("tabs");
    }
    
    void updateCharacterAttributes(MutableAttributeSet set, AttributeSet set2, boolean b) throws IOException {
        if (b) {
            Object attribute = set.getAttribute("characterStyle");
            Integer styleNumber = this.findStyleNumber(set2, "character");
            if (attribute != styleNumber) {
                if (attribute != null) {
                    this.resetCharacterAttributes(set);
                }
                if (styleNumber != null) {
                    this.writeControlWord("cs", (int)styleNumber);
                    set.addAttribute("characterStyle", styleNumber);
                }
                else {
                    set.removeAttribute("characterStyle");
                }
            }
        }
        Object attrDiff;
        if ((attrDiff = attrDiff(set, set2, StyleConstants.FontFamily, null)) != null) {
            this.writeControlWord("f", (int)this.fontTable.get(attrDiff));
        }
        this.checkNumericControlWord(set, set2, StyleConstants.FontSize, "fs", 12.0f, 2.0f);
        this.checkControlWords(set, set2, RTFAttributes.attributes, 0);
        this.checkNumericControlWord(set, set2, StyleConstants.LineSpacing, "sl", 0.0f, 20.0f);
        Object attrDiff2;
        if ((attrDiff2 = attrDiff(set, set2, StyleConstants.Background, RTFGenerator.MagicToken)) != null) {
            int intValue;
            if (attrDiff2 == RTFGenerator.MagicToken) {
                intValue = 0;
            }
            else {
                intValue = (int)this.colorTable.get(attrDiff2);
            }
            this.writeControlWord("cb", intValue);
        }
        Object attrDiff3;
        if ((attrDiff3 = attrDiff(set, set2, StyleConstants.Foreground, null)) != null) {
            int intValue2;
            if (attrDiff3 == RTFGenerator.MagicToken) {
                intValue2 = 0;
            }
            else {
                intValue2 = (int)this.colorTable.get(attrDiff3);
            }
            this.writeControlWord("cf", intValue2);
        }
    }
    
    protected void resetCharacterAttributes(MutableAttributeSet set) throws IOException {
        this.writeControlWord("plain");
        for (int length = RTFAttributes.attributes.length, i = 0; i < length; ++i) {
            RTFAttribute rtfAttribute = RTFAttributes.attributes[i];
            if (rtfAttribute.domain() == 0) {
                rtfAttribute.setDefault(set);
            }
        }
        StyleConstants.setFontFamily(set, "Helvetica");
        set.removeAttribute(StyleConstants.FontSize);
        set.removeAttribute(StyleConstants.Background);
        set.removeAttribute(StyleConstants.Foreground);
        set.removeAttribute(StyleConstants.LineSpacing);
        set.removeAttribute("characterStyle");
    }
    
    public void writeTextElement(Element element) throws IOException {
        this.updateCharacterAttributes(this.outputAttributes, element.getAttributes(), true);
        if (element.isLeaf()) {
            try {
                element.getDocument().getText(element.getStartOffset(), element.getEndOffset() - element.getStartOffset(), this.workingSegment);
            }
            catch (BadLocationException ex) {
                ex.printStackTrace();
                throw new InternalError(ex.getMessage());
            }
            this.writeText(this.workingSegment);
        }
        else {
            for (int elementCount = element.getElementCount(), i = 0; i < elementCount; ++i) {
                this.writeTextElement(element.getElement(i));
            }
        }
    }
    
    public void writeText(Segment segment) throws IOException {
        int i = segment.offset;
        int n = i + segment.count;
        char[] array = segment.array;
        while (i < n) {
            this.writeCharacter(array[i]);
            ++i;
        }
    }
    
    public void writeText(String s) throws IOException {
        for (int i = 0; i < s.length(); ++i) {
            this.writeCharacter(s.charAt(i));
        }
    }
    
    public void writeRawString(String s) throws IOException {
        for (int length = s.length(), i = 0; i < length; ++i) {
            this.outputStream.write((int)s.charAt(i));
        }
    }
    
    public void writeControlWord(String s) throws IOException {
        this.outputStream.write(92);
        this.writeRawString(s);
        this.afterKeyword = true;
    }
    
    public void writeControlWord(String s, int i) throws IOException {
        this.outputStream.write(92);
        this.writeRawString(s);
        this.writeRawString(String.valueOf(i));
        this.afterKeyword = true;
    }
    
    public void writeBegingroup() throws IOException {
        this.outputStream.write(123);
        this.afterKeyword = false;
    }
    
    public void writeEndgroup() throws IOException {
        this.outputStream.write(125);
        this.afterKeyword = false;
    }
    
    public void writeCharacter(char c) throws IOException {
        if (c == ' ') {
            this.outputStream.write(92);
            this.outputStream.write(126);
            this.afterKeyword = false;
            return;
        }
        if (c == '\t') {
            this.writeControlWord("tab");
            return;
        }
        if (c == '\n' || c == '\r') {
            return;
        }
        int convertCharacter = convertCharacter(this.outputConversion, c);
        if (convertCharacter == 0) {
            int i = 0;
            while (i < RTFGenerator.textKeywords.length) {
                if (RTFGenerator.textKeywords[i].character == c) {
                    this.writeControlWord(RTFGenerator.textKeywords[i].keyword);
                    return;
                }
                ++i;
            }
            String approximationForUnicode = this.approximationForUnicode(c);
            if (approximationForUnicode.length() != this.unicodeCount) {
                this.writeControlWord("uc", this.unicodeCount = approximationForUnicode.length());
            }
            this.writeControlWord("u", (int)c);
            this.writeRawString(" ");
            this.writeRawString(approximationForUnicode);
            this.afterKeyword = false;
            return;
        }
        else {
            if (convertCharacter > 127) {
                this.outputStream.write(92);
                this.outputStream.write(39);
                this.outputStream.write((int)RTFGenerator.hexdigits[(convertCharacter & 0xF0) >>> 4]);
                this.outputStream.write((int)RTFGenerator.hexdigits[convertCharacter & 0xF]);
                this.afterKeyword = false;
                return;
            }
            switch (convertCharacter) {
                case 92:
                case 123:
                case 125: {
                    this.outputStream.write(92);
                    this.afterKeyword = false;
                    break;
                }
            }
            if (this.afterKeyword) {
                this.outputStream.write(32);
                this.afterKeyword = false;
            }
            this.outputStream.write(convertCharacter);
            return;
        }
    }
    
    String approximationForUnicode(char c) {
        return "?";
    }
    
    static int[] outputConversionFromTranslationTable(char[] array) {
        int[] array2 = new int[2 * array.length];
        for (int i = 0; i < array.length; ++i) {
            array2[i * 2] = array[i];
            array2[i * 2 + 1] = i;
        }
        return array2;
    }
    
    static int[] outputConversionForName(String s) throws IOException {
        return outputConversionFromTranslationTable((char[])(char[])RTFReader.getCharacterSet(s));
    }
    
    protected static int convertCharacter(int[] array, char c) {
        int i = 0;
        while (i < array.length) {
            if (array[i] == c) {
                return array[i + 1];
            }
            i += 2;
        }
        return 0;
    }
    
    static {
        defaultRTFColor = Color.black;
        MagicToken = new Object();
        Dictionary<Object, Object> textKeywords = RTFReader.textKeywords;
        Enumeration<Object> keys = textKeywords.keys();
        Vector<Object> vector = new Vector();
        while (keys.hasMoreElements()) {
            RTFGenerator.CharacterKeywordPair obj = new RTFGenerator.CharacterKeywordPair();
            obj.keyword = (String)keys.nextElement();
            obj.character = ((String)textKeywords.get(obj.keyword)).charAt(0);
            vector.addElement(obj);
        }
        vector.copyInto(RTFGenerator.textKeywords = new RTFGenerator.CharacterKeywordPair[vector.size()]);
        hexdigits = new char[] { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };
    }
    
    static class CharacterKeywordPair
    {
        public char character;
        public String keyword;
    }
}
